from enum import Enum
from typing import Union, List, Set, Any

from pydantic import BaseModel

from guniflask.orm.base_model import BaseModelMixin
from guniflask.utils.rule import make_ignore_rule_for_field, make_only_rule_for_field


class DataModel(BaseModel):
    def to_dict(
            self,
            ignore: Union[str, List[str], Set[str]] = None,
            only: Union[str, List[str], Set[str]] = None,
    ) -> dict:
        ignore = make_ignore_rule_for_field(ignore)
        only = make_only_rule_for_field(only)
        return self.__to_dict(ignore, only)

    def __to_dict(
            self,
            ignore: Set[str],
            only: Set[str],
            _prefix: str = '',
    ) -> dict:
        data = {}
        for k, v in self.__dict__.items():
            if _in_set(_prefix, k, ignore):
                continue
            if only and not _in_set(_prefix, k, only):
                continue
            data[k] = self.__get_value_from(
                v,
                ignore=ignore,
                only=only,
                _prefix=_new_prefix(_prefix, k),
            )
        return data

    def __get_value_from(
            self,
            v: Any,
            ignore: Set[str],
            only: Set[str],
            _prefix: str = '',
    ) -> Any:
        if isinstance(v, DataModel):
            return v.__to_dict(ignore=ignore, only=only, _prefix=_prefix)
        if isinstance(v, dict):
            data = {}
            for _k, _v in v.items():
                if _in_set(_prefix, _k, ignore):
                    continue
                if only and not _in_set(_prefix, _k, only):
                    continue
                data[_k] = self.__get_value_from(
                    _v,
                    ignore=ignore,
                    only=only,
                    _prefix=_new_prefix(_prefix, _k),
                )
            return data
        if isinstance(v, (list, set, tuple)):
            return [
                self.__get_value_from(
                    _v,
                    ignore=ignore,
                    only=only,
                    _prefix=_prefix,
                ) for _v in v
            ]
        if isinstance(v, Enum):
            return v.value
        return v

    @classmethod
    def from_dict(cls, obj: dict):
        return cls(**obj)

    @classmethod
    def from_orm(
            cls,
            obj: Any,
            ignore: Union[str, List[str], Set[str]] = None,
            only: Union[str, List[str], Set[str]] = None,
            include: Union[str, List[str], Set[str]] = None,
    ):
        if isinstance(obj, BaseModelMixin):
            return cls.from_dict(
                obj.to_dict(
                    ignore=ignore,
                    only=only,
                    include=include,
                )
            )
        return super().from_orm(obj)


def _expand_rule_set(rule: Set[str]) -> dict:
    result = {}
    for r in rule:
        s = r.split('.')
        m = result
        for i in s[:-1]:
            if i not in m:
                m[i] = {}
            m = m[i]
        if s[-1] not in m:
            m[s[-1]] = None
    return result


def _make_ignore_rule(ignore: Union[str, List[str], Set[str]]) -> dict:
    return _expand_rule_set(make_ignore_rule_for_field(ignore))


def _make_only_rule(only: Union[str, List[str], Set[str]]) -> dict:
    return _expand_rule_set(make_only_rule_for_field(only))


def _in_set(prefix: str, key: str, s: set):
    return _new_prefix(prefix, key) in s


def _new_prefix(prefix: str, key: str):
    if not prefix:
        return key
    return prefix + '.' + key



from guniflask.beans.definition import BeanDefinition
from guniflask.beans.definition_registry import BeanDefinitionRegistry
from guniflask.context.autowired_post_processor import AutowiredAnnotationBeanPostProcessor
from guniflask.context.config_constants import *
from guniflask.context.config_post_processor import ConfigurationClassPostProcessor


class AnnotationConfigUtils:

    @staticmethod
    def register_annotation_config_processors(registry: BeanDefinitionRegistry):
        if not registry.contains_bean_definition(CONFIGURATION_ANNOTATION_PROCESSOR):
            bean_definition = BeanDefinition(ConfigurationClassPostProcessor)
            registry.register_bean_definition(CONFIGURATION_ANNOTATION_PROCESSOR,
                                              bean_definition)
        if not registry.contains_bean_definition(AUTOWIRED_ANNOTATION_PROCESSOR):
            bean_definition = BeanDefinition(AutowiredAnnotationBeanPostProcessor)
            registry.register_bean_definition(AUTOWIRED_ANNOTATION_PROCESSOR,
                                              bean_definition)



from .default_factory import DefaultBeanFactory
from .definition import BeanDefinition
from .definition_registry import BeanDefinitionRegistry
from .factory import BeanFactory
from .factory import BeanFactoryAware
from .factory import BeanNameAware
from .factory import ConfigurableBeanFactory
from .lifecycle import DisposableBean
from .lifecycle import InitializingBean
from .lifecycle import SmartInitializingSingleton
from .post_processor import BeanPostProcessor
from .singleton_registry import SingletonBeanRegistry



from typing import Optional

from guniflask.context.annotation import configuration, bean
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security_config.authentication_manager_builder import AuthenticationManagerBuilder


@configuration
class AuthenticationConfiguration:
    def __init__(self):
        self._authentication_manager: Optional[AuthenticationManager] = None
        self._authentication_manager_initialized = False
        self._authentication_manager_builder = AuthenticationManagerBuilder()

    @bean
    def authentication_manager_builder(self) -> AuthenticationManagerBuilder:
        return self._authentication_manager_builder

    @property
    def authentication_manager(self) -> AuthenticationManager:
        if self._authentication_manager_initialized:
            return self._authentication_manager
        self._authentication_manager = self._authentication_manager_builder.build()
        return self._authentication_manager



import datetime as dt

from pydantic.datetime_parse import parse_datetime

local_tz_info = dt.datetime.now().astimezone().tzinfo


def convert_to_datetime(s):
    return parse_datetime(s).astimezone()



import inspect
import logging
from functools import partial, update_wrapper

from guniflask.annotation import AnnotationUtils
from guniflask.beans.factory import BeanFactory, BeanFactoryAware
from guniflask.beans.lifecycle import SmartInitializingSingleton
from guniflask.beans.post_processor import BeanPostProcessor
from guniflask.scheduling.annotation import AsyncRun
from guniflask.scheduling.async_executor import AsyncExecutor

log = logging.getLogger(__name__)


class AsyncPostProcessor(BeanPostProcessor, BeanFactoryAware, SmartInitializingSingleton):

    def __init__(self):
        self.bean_factory = None
        self._async_methods = []

    def set_bean_factory(self, bean_factory: BeanFactory):
        self.bean_factory = bean_factory

    def post_process_after_initialization(self, bean, bean_name: str):
        bean_type = bean.__class__
        for m in dir(bean_type):
            func = getattr(bean_type, m)
            if inspect.isfunction(func):
                a = AnnotationUtils.get_annotation(func, AsyncRun)
                if a is not None:
                    method = getattr(bean, m)
                    self._async_methods.append((a, bean, method))
        return bean

    def after_singletons_instantiated(self):
        self._finish_registration()

    def _finish_registration(self):
        assert self.bean_factory is not None and isinstance(self.bean_factory, BeanFactory), \
            'Bean factory must be set to find async executors'
        for async_run, bean, method in self._async_methods:
            self._process_async(async_run, bean, method)

    def _process_async(self, async_run: AsyncRun, bean, method):
        if async_run['executor'] is None:
            async_executor = self.bean_factory.get_bean_of_type(AsyncExecutor)
        else:
            async_executor = self.bean_factory.get_bean(async_run['executor'], required_type=AsyncExecutor)
            if async_executor is None:
                log.error('Cannot find the async executor named "%s"', async_run['executor'])
        if async_executor is not None:
            method_name = method.__name__
            wrapped_method = self.wrap_async_func(method, async_executor)
            setattr(bean, method_name, wrapped_method)

    def wrap_async_func(self, func, async_executor):
        def wrapper(*args, **kwargs):
            wrapped_func = self._post_process_async_method(func)
            task = partial(wrapped_func, *args, **kwargs)
            async_executor.submit(task)

        return update_wrapper(wrapper, func)

    def _post_process_async_method(self, method):
        return method



from guniflask.context.annotation import bean, configuration
from guniflask.scheduling.async_config import AsyncConfiguration
from guniflask.scheduling.async_post_processor import AsyncPostProcessor
from guniflask.scheduling.config_constants import ASYNC_ANNOTATION_PROCESSOR, SCHEDULED_ANNOTATION_PROCESSOR
from guniflask.scheduling.scheduling_config import SchedulingConfiguration
from guniflask.scheduling.scheduling_post_processor import ScheduledPostProcessor
from guniflask.utils.context import run_with_context


@configuration
class WebAsyncConfiguration(AsyncConfiguration):

    @bean(ASYNC_ANNOTATION_PROCESSOR)
    def async_annotation_processor(self) -> AsyncPostProcessor:
        return WebAsyncPostProcessor()


class WebAsyncPostProcessor(AsyncPostProcessor):

    def _post_process_async_method(self, method):
        wrapped_method = run_with_context(method)
        return super()._post_process_async_method(wrapped_method)


@configuration
class WebSchedulingConfiguration(SchedulingConfiguration):

    @bean(SCHEDULED_ANNOTATION_PROCESSOR)
    def scheduled_annotation_processor(self) -> ScheduledPostProcessor:
        return WebScheduledPostProcessor()


class WebScheduledPostProcessor(ScheduledPostProcessor):

    def _post_process_scheduled_method(self, method):
        wrapped_method = run_with_context(method)
        return super()._post_process_scheduled_method(wrapped_method)



import os
from importlib import import_module
from os.path import isfile, join, isdir
from pkgutil import iter_modules


def walk_modules(path: str):
    mods = []
    mod = import_module(path)
    mods.append(mod)
    if hasattr(mod, '__path__'):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + '.' + subpath
            if ispkg:
                mods += walk_modules(fullpath)
            else:
                submod = import_module(fullpath)
                mods.append(submod)
    return mods


def walk_files(path: str):
    files = []
    if isdir(path):
        names = os.listdir(path)
        for name in names:
            files += walk_files(join(path, name))
    elif isfile(path):
        files.append(path)
    return files



from typing import Type

from guniflask.beans.factory import BeanFactory
from guniflask.context.event import ApplicationEvent
from guniflask.context.event_listener import ApplicationEventListener
from guniflask.data_model.typing import inspect_args


class ApplicationEventPublisher:
    def __init__(self, bean_factory: BeanFactory):
        self._bean_factory = bean_factory
        self._app_listeners = set()
        self._app_listener_beans = set()

    def add_application_listener(self, listener: ApplicationEventListener):
        self._app_listeners.add(listener)

    def add_application_listener_bean(self, bean_name: str):
        self._app_listener_beans.add(bean_name)

    def publish_event(self, event: ApplicationEvent):
        for listener in self._get_application_listeners():
            assert isinstance(listener, ApplicationEventListener)
            accepted_event_type = self._resolve_accepted_event_type(listener.on_application_event)
            if accepted_event_type is None or isinstance(event, accepted_event_type):
                listener.on_application_event(event)

    def _get_application_listeners(self):
        for listener in self._app_listeners:
            yield listener
        for bean_name in self._app_listener_beans:
            listener = self._bean_factory.get_bean(bean_name, required_type=ApplicationEventListener)
            if listener is not None:
                yield listener

    def _resolve_accepted_event_type(self, method) -> Type[ApplicationEvent]:
        args, hints = inspect_args(method)
        event_type_arg = None
        for k in args:
            event_type_arg = k
            break
        if event_type_arg is not None:
            return hints.get(event_type_arg)



import datetime as dt
import inspect
import logging
from functools import update_wrapper
from typing import Optional

from flask import Blueprint as FlaskBlueprint, request, current_app
from pydantic.main import BaseModel

from guniflask.annotation import AnnotationUtils
from guniflask.beans.post_processor import BeanPostProcessor
from guniflask.context.event import ContextRefreshedEvent, ApplicationEvent
from guniflask.context.event_listener import ApplicationEventListener
from guniflask.data_model.typing import parse_json, inspect_args, analyze_arg_type
from guniflask.utils.datatime import convert_to_datetime
from guniflask.web import request_annotation
from guniflask.web.bind_annotation import Blueprint, Route
from guniflask.web.errors import RequestValidationError
from guniflask.web.filter_annotation import MethodDefFilter
from guniflask.web.request_annotation import FieldInfo
from guniflask.web.request_annotation import FilePartInfo, FormValueInfo, RequestHeaderInfo, CookieValueInfo
from guniflask.web.request_annotation import PathVariableInfo
from guniflask.web.request_annotation import RequestBody, RequestBodyInfo
from guniflask.web.request_annotation import RequestParam, RequestParamInfo

log = logging.getLogger(__name__)


class BlueprintPostProcessor(BeanPostProcessor, ApplicationEventListener):
    def __init__(self):
        self.blueprints = []

    def on_application_event(self, application_event: ApplicationEvent):
        if isinstance(application_event, ContextRefreshedEvent):
            self._register_blueprints()

    def post_process_after_initialization(self, bean, bean_name: str):
        bean_type = bean.__class__
        annotation = AnnotationUtils.get_annotation(bean_type, Blueprint)
        if annotation is not None:
            options = annotation['options'] or {}
            b = FlaskBlueprint(bean_name, bean.__module__,
                               url_prefix=annotation['url_prefix'], **options)
            annotation['blueprint'] = b
            for m in dir(bean_type):
                func = getattr(bean_type, m)
                if inspect.isfunction(func):
                    method = getattr(bean, m)
                    try:
                        self._resolve_route(b, method)
                        self._resolve_method_def_filter(b, method)
                    except Exception:
                        log.error(
                            "Failed to resolve the route function '%s' in blueprint '%s'",
                            m,
                            bean_type.__name__,
                        )
                        raise
            self.blueprints.append(b)
        return bean

    def _resolve_route(self, blueprint: FlaskBlueprint, method):
        a = AnnotationUtils.get_annotation(method, Route)
        if a is not None:
            rule_options = a['options'] or {}
            blueprint.add_url_rule(
                a['rule'],
                endpoint=method.__name__,
                view_func=self._wrap_view_func(a['rule'], method),
                **rule_options
            )

    def _wrap_view_func(self, rule: str, method):
        params, param_names = self._resolve_method_parameters(rule, method)

        def wrapper(**kwargs):
            try:
                method_kwargs = self._resolve_method_kwargs(kwargs, params, param_names)
            except Exception as e:
                raise RequestValidationError(e)
            return method(**method_kwargs)

        return update_wrapper(wrapper, method)

    def _resolve_method_def_filter(self, blueprint, method):
        a = AnnotationUtils.get_annotation(method, MethodDefFilter)
        if a is None:
            return
        for v in a['values']:
            name = v['name']
            args = v['args']
            if args is None:
                getattr(blueprint, name)(method)
            else:
                getattr(blueprint, name)(*args)(method)

    def _register_blueprints(self):
        for b in self.blueprints:
            current_app.register_blueprint(b)

    def _resolve_method_parameters(self, rule: str, method):
        """
        :param rule: path rule
        :param method: view function
        """
        params = {}
        param_names = {}

        args, type_hints = inspect_args(method)

        for arg, default in args.items():
            if default is inspect._empty:
                required = True
                default = None
            else:
                required = False
            if inspect.isfunction(default) and getattr(request_annotation, default.__name__, None) is default:
                default = default()
            if not isinstance(default, FieldInfo):
                if arg in type_hints:
                    arg_type = type_hints[arg]
                    arg_ = analyze_arg_type(arg_type)
                    if arg_.is_dict():
                        annotation = RequestBody()
                    else:
                        annotation = RequestParam()
                else:
                    annotation = RequestParam()
                annotation.default = default
                default = annotation
            if arg in type_hints and default.dtype is None:
                default.dtype = type_hints[arg]
            if default.required is None:
                default.required = required
            params[arg] = default
            if default.name is not None:
                param_names[default.name] = arg

        return params, param_names

    def _resolve_method_kwargs(self, kwargs: dict, params: dict, param_names: dict) -> dict:
        """
        :param kwargs: kwargs injected by Flask
        :param params: view function parameters
        :param param_names: names of the above parameters
        """
        result = {}
        for k, v in kwargs.items():
            if k in param_names:
                k = param_names[k]
            if k not in params:
                raise AssertionError(f'Unhandled parameter: {k}={v}')
            if params[k].dtype is not None:
                try:
                    v = params[k].dtype(v)
                except ValueError:
                    pass
            result[k] = v
        for k, p in params.items():
            if k in result:
                continue
            name = p.name or k
            if isinstance(p, PathVariableInfo):
                raise AssertionError(f'No such path variable: {name}')

            if isinstance(p, RequestParamInfo):
                if inspect.isclass(p.dtype) and issubclass(p.dtype, BaseModel):
                    v = parse_json(request.args, dtype=p.dtype)
                    result[k] = v
                else:
                    if name in request.args:
                        arg_ = analyze_arg_type(p.dtype)
                        if arg_.is_list():
                            v = request.args.getlist(name)
                            if arg_.outer_type:
                                for i in range(len(v)):
                                    v[i] = self._read_value(v[i], arg_.outer_type)
                        elif arg_.is_singleton():
                            v = request.args.get(name)
                            if p.dtype is not None:
                                v = self._read_value(v, p.dtype)
                        else:
                            raise AssertionError(f'Unsupported type of parameter "{name}": {p.dtype}')
                        if v is not None:
                            result[k] = v
            elif isinstance(p, RequestBodyInfo):
                if inspect.isclass(p.dtype) and issubclass(p.dtype, bytes):
                    result[k] = request.data
                else:
                    v = parse_json(request.json, dtype=p.dtype)
                    if v is not None:
                        result[k] = v
            elif isinstance(p, FilePartInfo):
                file = request.files.get(name)
                if file is not None:
                    if p.dtype == bytes:
                        result[k] = file.read()
                    else:
                        result[k] = file
            elif isinstance(p, FormValueInfo):
                arg_ = analyze_arg_type(p.dtype)
                if arg_.is_list():
                    v = request.form.getlist(name)
                    if arg_.outer_type:
                        for i in range(len(v)):
                            v[i] = self._read_value(v[i], arg_.outer_type)
                elif arg_.is_singleton():
                    v = request.form.get(name)
                    if p.dtype is not None:
                        v = self._read_value(v, p.dtype)
                else:
                    raise AssertionError(f'Unsupported type of parameter "{name}": {p.dtype}')
                if v is not None:
                    result[k] = v
            elif isinstance(p, RequestHeaderInfo):
                arg_ = analyze_arg_type(p.dtype)
                if arg_.is_list():
                    v = request.headers.getlist(name)
                    if arg_.outer_type:
                        for i in range(len(v)):
                            v[i] = self._read_value(v[i], arg_.outer_type)
                elif arg_.is_singleton():
                    v = request.headers.get(name)
                    if p.dtype is not None:
                        v = self._read_value(v, p.dtype)
                else:
                    raise AssertionError(f'Unsupported type of header "{name}": {p.dtype}')
                if v is not None:
                    result[k] = v
                v = self._read_value(request.headers.get(name), p.dtype)
                if v is not None:
                    result[k] = v
            elif isinstance(p, CookieValueInfo):
                v = request.cookies.get(name)
                if v is not None:
                    if p.dtype is not None:
                        v = self._read_value(v, p.dtype)
                    result[k] = v

            if k not in result:
                if p.required:
                    if isinstance(p, RequestBodyInfo):
                        raise AssertionError('Request body not given or in wrong format')
                    else:
                        raise AssertionError(f'Parameter not given: {name}')
                result[k] = p.default
        return result

    def _read_value(self, v: Optional[str], dtype: type):
        if v is None:
            return
        if dtype == bool:
            try:
                return bool(int(v))
            except (ValueError, TypeError):
                if v in ("True", "true"):
                    return True
                if v in ("False", "false"):
                    return False
        elif dtype == dt.datetime:
            return convert_to_datetime(v)
        else:
            try:
                return dtype(v)
            except (ValueError, TypeError):
                raise AssertionError(f'Failed to read value, type: {dtype.__name__}, value: {v}')



import inspect
from typing import List, Set, Mapping, Dict

import pytest

from guniflask.data_model.typing import parse_json, analyze_arg_type, inspect_args


class Person:
    name: str
    age: int


class Teacher(Person):
    classes: Set


class Student(Person):
    secret: str
    mentor: Teacher
    parents: List[Person]
    scores: dict
    hobbies: List = None
    graduated: bool = False


def test_parse_simple_data():
    assert parse_json('1', dtype=int) == 1
    assert parse_json('1', dtype=List[int]) == [1]
    assert parse_json([1, 2, 3]) == [1, 2, 3]
    assert parse_json({'key': 'value'}) == {'key': 'value'}


def test_parse_object_data():
    student_data = dict(
        name='Bob',
        age=12,
        graduated=True,
        hobbies=['Programming', 'Piano'],
        scores={'Math': 100},
        parents=[
            dict(
                name='Billy',
                age=40
            ),
            dict(
                name='Judy',
                age=39
            )
        ],
        mentor=dict(
            name='Alice',
            age=41,
            classes=['English', 'Math']
        )
    )

    student = parse_json(student_data, Student)

    assert isinstance(student, Student)
    assert student.name == 'Bob'
    assert student.age == 12
    assert student.graduated is True
    assert student.hobbies == ['Programming', 'Piano']
    assert student.scores == {'Math': 100}
    assert not hasattr(student, 'secret')

    assert isinstance(student.parents, list) and len(student.parents) == 2
    for parent in student.parents:
        if parent.name == 'Billy':
            assert parent.age == 40
        elif parent.name == 'Judy':
            assert parent.age == 39
        else:
            raise RuntimeError

    assert isinstance(student.mentor, Teacher)
    mentor = student.mentor
    assert mentor.name == 'Alice'
    assert mentor.age == 41
    assert isinstance(mentor.classes, set)
    assert mentor.classes == {'English', 'Math'}


def test_parse_list():
    class Prop:
        name: str
        value: str

    prop_list = parse_json({'name': 'name', 'value': 'Alice'}, List[Prop])
    assert isinstance(prop_list, list) and len(prop_list) == 1
    prop = prop_list[0]
    assert prop.name == 'name' and prop.value == 'Alice'

    prop_list = parse_json(None, List[Prop])
    assert prop_list is None


def test_analyze_arg_type():
    arg_ = analyze_arg_type(None)
    assert arg_.is_singleton() and arg_.outer_type is None

    with pytest.raises(AssertionError):
        analyze_arg_type('')

    arg_ = analyze_arg_type(list)
    assert arg_.is_list() and arg_.outer_type is None

    arg_ = analyze_arg_type(set)
    assert arg_.is_set() and arg_.outer_type is None

    arg_ = analyze_arg_type(dict)
    assert arg_.is_dict() and arg_.outer_type == (None, None)

    arg_ = analyze_arg_type(int)
    assert arg_.is_singleton() and arg_.outer_type is int

    arg_ = analyze_arg_type(str)
    assert arg_.is_singleton() and arg_.outer_type is str

    arg_ = analyze_arg_type(List)
    assert arg_.is_list() and arg_.outer_type is None

    arg_ = analyze_arg_type(Set)
    assert arg_.is_set() and arg_.outer_type is None

    arg_ = analyze_arg_type(Mapping)
    assert arg_.is_dict() and arg_.outer_type == (None, None)

    arg_ = analyze_arg_type(Dict)
    assert arg_.is_dict() and arg_.outer_type == (None, None)

    class A:
        pass

    arg_ = analyze_arg_type(A)
    assert arg_.is_singleton() and arg_.outer_type is A

    arg_ = analyze_arg_type(List[A])
    assert arg_.is_list() and arg_.outer_type is A
    arg_ = analyze_arg_type(List[str])
    assert arg_.is_list() and arg_.outer_type is str

    arg_ = analyze_arg_type(Set[A])
    assert arg_.is_set() and arg_.outer_type is A
    arg_ = analyze_arg_type(Set[str])
    assert arg_.is_set() and arg_.outer_type is str

    arg_ = analyze_arg_type(Mapping[str, A])
    assert arg_.is_dict() and arg_.outer_type == (str, A)
    arg_ = analyze_arg_type(Mapping[str, str])
    assert arg_.is_dict() and arg_.outer_type == (str, str)

    arg_ = analyze_arg_type(Dict[str, A])
    assert arg_.is_dict() and arg_.outer_type == (str, A)
    arg_ = analyze_arg_type(Dict[str, str])
    assert arg_.is_dict() and arg_.outer_type == (str, str)

    arg_ = analyze_arg_type(Mapping[str, List[A]])
    assert arg_.is_dict() and type(arg_.outer_type) == tuple
    outer_type = arg_.outer_type
    assert outer_type[0] == str
    assert outer_type[1].is_list() and outer_type[1].outer_type == A


def test_inspect_args():
    class A:
        pass

    def func(a, b: int, c: List, d=1, e: List = None, f: List[str] = None, g: A = None) -> dict:
        pass

    args, hints = inspect_args(func)
    assert args['a'] is inspect._empty
    assert args['b'] is inspect._empty
    assert args['c'] is inspect._empty
    assert args['d'] == 1
    assert args['e'] is None
    assert args['f'] is None
    assert args['g'] is None
    assert hints == {'return': dict, 'b': int, 'c': List, 'e': List, 'f': List[str], 'g': A}
